/*
 * Very simple framebuffer demo for GPGPU.
 *
 * It performs a number of times the 'saxpy' operation, i.e.,
 *   y_{i+1} = a * x + y_i
 *
 * Author: Dominik Goeddeke
 *
 * Modifications: Gabriel Zachmann, June 2007;
 *
 * I have removed a lot of the stuff that deals with different variations, and
 * older parts of the OpenGL API. In fact, you probably won't recognize the original
 * source code ;-)
 */



// includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#if defined(__linux__)

#define GL_GLEXT_PROTOTYPES
#include <gl.h>
#include <glu.h>
#include <glut.h>

#elif defined(__APPLE__)

#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>

#endif

#include "shader_setup.h"


// prototypes
bool checkFramebufferStatus(void);
void checkGLErrors(const char *label);
void compareResults(void);
void createTextures(void);
void initGLSL(void);
void initFBO(void);
void performComputation(void);
void setupTexture(const GLuint texID);
void swapTextures(void);
void transferFromTexture(float* data);
void transferToTexture(float* data, GLuint texID);
void changeSize(int w, int h);
void renderScene(void);
void processNormalKeys(unsigned char key, int x, int y);


// command line options
unsigned int N = 100;							// length of saxpy operands
unsigned int texSizeX, texSizeY;		// derived from N
int numIterations = 100;				// num iterations for timing
GLenum texInternalFormat = GL_RGBA32F_ARB;		// default = 32 bit internal texture format
bool with_window = false;

unsigned int windowWidth = 0, windowHeight = 0;

// flags to fine-tune application and to ease debugging
bool showResults = false;

// support of extensions
bool npot_supported;

// texture identifiers
GLuint yTexID[2];
GLuint xTexID;

// ping pong management vars
int writeTex = 0;
int readTex = 1;
GLenum attachmentpoints[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };

// GLSL vars
GLuint sh_prog_id;
GLint yUni, xUni, alphaUni;

// FBO identifier
GLuint fbo;

// actual data
float* dataX;
float* dataY;
float alpha;


/**
 * main, just calls things in the appropriate order
 */

int main(int argc, char **argv)
{

	// Sets up GLUT, creates "window"
	// (better put: valid GL context, since the window is not really needed --
	// it's only needed with option -w)
	glutInit( &argc, argv );
	// create window, if wanted
	if ( with_window )
	{
		glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB );
		glutInitWindowPosition(100,100);
		glutInitWindowSize( 300, 300 );
	}
	glutCreateWindow("saxpy test");		// must be done before any other GL commands

	npot_supported = static_cast<bool>( glutExtensionSupported("GL_ARB_texture_non_power_of_two") );

	// parse command line
	for ( int i = 1; i < argc; i ++ )
	{
		if ( argv[i][0] == '-' )
		{
			if ( strcmp(argv[i],"-16") == 0 )
				texInternalFormat = GL_RGBA16F_ARB;
			else
			if ( strcmp(argv[i],"-v") == 0 )
				showResults = true;
			else
			if ( strcmp(argv[i],"-w") == 0 )
				with_window = true;
			else
			{
				puts( "saxpy [options] [array_size num_iter]\n"
					  "-16     use 16 bits for internal texture (i.e., float) format\n"
					  "-v      print result vectors (both GPU and CPU)\n"
					  "-w      show final texture (= array) in a window\n"
					  "-h      this help\n" );
				exit(0);
			}

		}
		else
		{
			// expecting 2 numerical args
			if ( argc - i < 2 )
			{
				fputs("Expecting 2 arguments: array_size AND num_iter!\n", stderr );
				exit(1);
			}

			N = atoi( argv[i] );
			if ( N < 4 )
			{
				fprintf(stderr, "N (%d) < 4 !\n", N );
				N = 4;
			}

			i += 1;
			numIterations = atoi( argv[i] );
			if ( numIterations < 1 )
				numIterations = 1;
		}
	}

	// calc texture dimensions
	if ( (N/4)*4 != N )
	{
		fprintf( stderr, "N (%u) not divisible by 4!\n", N );
		exit(1);
	}

	texSizeX = static_cast<unsigned int>( sqrt( N/4.0f ) );
	texSizeY = (N/4) / texSizeX;
	int i = texSizeX/2;
	while ( (N/4) != texSizeX*texSizeY && i > 0 )
	{
		texSizeX -- ;
		texSizeY = (N/4) / texSizeX;
		i -- ;
	}
	if ( (N/4) != texSizeX*texSizeY )
	{
		fprintf( stderr, "texSizeX * texSizeY (%u * %u) != N/4 (%u) !\n",
				 texSizeX, texSizeY, N/4 );
		exit(1);
	}

	// check whether we can actually load textures of that size on the GPU
	glTexImage2D( GL_PROXY_TEXTURE_2D, 0, texInternalFormat,
				  texSizeX, texSizeY, 0, GL_RGBA, GL_FLOAT, NULL );
	GLint realwidth;
	glGetTexLevelParameteriv( GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &realwidth );
	if ( realwidth == 0 )
	{
		fprintf( stderr, "Can't load textures of size %u x %u on the graphics card!\n",
				 texSizeX, texSizeY );
		exit(1);
	}

	// check if we've got a NPOT texture, and if that is supported (if any)
	if ( texSizeX != static_cast<unsigned int>( exp2( round( log2(texSizeX) ) ) ) ||
	     texSizeY != static_cast<unsigned int>( exp2( round( log2(texSizeY) ) ) )    )
	{
		fprintf( stderr, "texture size is  %u x %u x RGBA\n", texSizeX, texSizeY );

		if ( npot_supported )
			fputs( "NPOT textures are OK.\n", stderr );
		else
		{
			fputs( "NPOT textures are NOT supported!\n", stderr );
			exit(1);
		}
	}

    // create data vectors
    dataX = (float*) malloc( N*sizeof(float) );
	dataY = (float*) malloc( N*sizeof(float) );

    // and fill with some arbitrary values
	for (unsigned int i=0; i<N; i++)
	{
		dataX[i] = 2.0;
		dataY[i] = i+1.0;
	}
	alpha = 1.0/9.0;

	initFBO();							// init off-screen framebuffer
	createTextures();					// create textures for vectors
	initGLSL();							// load shaders
    performComputation();				// and start computation
	compareResults();					// compare results

	if ( with_window )
	{
		// show final result, which is in the "read" texture

		glutDisplayFunc( renderScene );
		glutReshapeFunc( changeSize );
		glutKeyboardFunc( processNormalKeys );

		glUseProgram( 0 );				// use fixed-function pipeline

		// scale texels into [0,1]
		float * data = new float[N];
		transferFromTexture( data );

		float maxi = -1E20, mini = 1E20;
		for ( unsigned int i = 0; i < N; i ++ )
		{
			if ( data[i] > maxi )
				maxi = data[i];
			if ( data[i] < mini )
				mini = data[i];
		}
		for ( unsigned int i = 0; i < N; i ++ )
			data[i] = (data[i] - mini) / (maxi - mini);

		transferToTexture( data, yTexID[readTex] );		// this texture is used by renderScene()
		delete [] data;

		// now show it
		glutMainLoop();					// this will never return
	}

    glDeleteFramebuffersEXT( 1, &fbo );		// and clean up

    free(dataX);
    free(dataY);
    glDeleteTextures( 2, yTexID );
    glDeleteTextures( 1, &xTexID );

    return 0;							// exit
}


/**
 * Sets up a floating point texture with NEAREST filtering.
 * (mipmaps etc. are unsupported for floating point textures)
 */
void setupTexture( const GLuint texID )
{
	// make active and bind
	glBindTexture( GL_TEXTURE_2D, texID );

	// turn off filtering and wrap modes!
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	// define texture with floating point format
	glTexImage2D( GL_TEXTURE_2D, 0, texInternalFormat,
				  texSizeX, texSizeY, 0, GL_RGBA, GL_FLOAT, NULL );

	// check if that worked
	checkGLErrors("setupTexture");
}


/**
 * Transfers data from current "read" texture, and stores it in given array.
 */
void transferFromTexture(float* data)
{
	// version a: texture is attached
	// recommended on both NVIDIA and ATI
    glReadBuffer( attachmentpoints[readTex] );
	glReadPixels( 0, 0, texSizeX, texSizeY, GL_RGBA, GL_FLOAT, data );

	// version b: texture is not neccessarily attached
//	glBindTexture(GL_TEXTURE_2D,yTexID[readTex]);
//	glGetTexImage(GL_TEXTURE_2D,0,GL_RGBA,GL_FLOAT,data);
}


/**
 * Transfers data to texture. 
 * Check web page for detailed explanation on the difference between ATI and NVIDIA.
 */
void transferToTexture( float* data, GLuint texID )
{
	// version (a): HW-accelerated on NVIDIA 
	glBindTexture( GL_TEXTURE_2D, texID);
	glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, texSizeX, texSizeY,
					 GL_RGBA, GL_FLOAT, data );

	// version (b): HW-accelerated on ATI 
//	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, texID, 0);
//	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
//	glRasterPos2i(0,0);
//	glDrawPixels(texSize,texSize,GL_RGBA,GL_FLOAT,data);
//	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, 0, 0);
}


/**
 * creates textures
 */
void createTextures(void)
{
    // create textures 
    // y gets two textures, alternatingly read-only and write-only, 
    // x is just read-only
    glGenTextures( 2, yTexID );
    glGenTextures( 1, &xTexID );

    // set up textures
    setupTexture( yTexID[readTex] );
	transferToTexture( dataY, yTexID[readTex] );
    setupTexture( yTexID[writeTex] );
	transferToTexture( dataY, yTexID[writeTex] );
    setupTexture( xTexID );
	transferToTexture( dataX, xTexID );

	// set texenv mode from modulate (the default) to replace)
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );

    // check if something went completely wrong
    checkGLErrors("createTextures()");
}



/**
 * Create framebuffer object, bind it to reroute rendering operations 
 * from the traditional framebuffer to the off-screen buffer
 */
void initFBO(void)
{
	// create FBO (off-screen framebuffer)
    glGenFramebuffersEXT( 1, &fbo ); 

    // bind offscreen framebuffer (that is, skip the window-specific render target)
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fbo );

	// create 'viewport' exactly with same dimensions as the texture(s)
    glViewport( 0, 0, texSizeX, texSizeY );

    // orthographic projection for 1:1 pixel=texture mapping
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluOrtho2D( 0.0, texSizeX, 0.0, texSizeY );
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // check if something went completely wrong
    checkGLErrors("initFBO()");
}


/**
 * Sets up the GLSL runtime and creates shader.
 */
void initGLSL(void)
{
	sh_prog_id = setShaders( "saxpy.vert", "saxpy.frag" );

	// Get location of the texture samplers for future use
    yUni = glGetUniformLocation( sh_prog_id, "textureY" );
	xUni = glGetUniformLocation( sh_prog_id, "textureX" );
	alphaUni = glGetUniformLocation( sh_prog_id, "alpha" );
}


/**
 * Performs and times saxpy on the CPU, compares results
 */
void compareResults(void)
{
    // get GPU results
    float* data = new float[N];
    transferFromTexture( data );

	// compute "ground truth" using double's
	double *dataYD = new double[N];
	double *dataXD = new double[N];
	for ( unsigned int i = 0; i < N; i ++ ) 
	{
		dataYD[i] = static_cast<double>( dataY[i] );
		dataXD[i] = static_cast<double>( dataX[i] );
	}

	// calc on CPU
	double start = clock();
	for (int n = 0; n < numIterations; n ++ ) 
		for ( unsigned int i = 0; i < N; i ++ ) 
			dataYD[i] += alpha*dataXD[i];
	double end = clock();
	double total = (end-start)/CLOCKS_PER_SEC;
	double mflops = (2.0*N*numIterations) / (total * 1000000.0);
	printf("CPU MFLOPs: %f\n", mflops );

	// and compare results
	double maxError = -1000.0;
	double avgError = 0.0;
	for ( unsigned int i = 0; i < N; i ++ )
	{
		double diff = fabs(data[i] - dataYD[i]) / fabs(dataYD[i]);
		if ( diff > maxError )
			maxError = diff;
		avgError += diff;
	}

	avgError /= static_cast<double>( N );
	printf("Max error: %e  (rel. to the absolute value)\n", maxError );
	printf("Avg error: %e  (rel. to the absolute value)\n", avgError );

	if ( showResults )
	{
		printf("CPU results:       GPU results:\n");

		for ( unsigned int i = 0; i < N; i ++ ) 
			printf("%10f       %10f\n", dataYD[i], data[i] );
	}

	delete [] data;
	delete [] dataYD;
	delete [] dataXD;
}


/**
 * Performs the actual calculation.
 */
void performComputation(void)
{
	// attach two textures to FBO
	// because it's much faster to select a new render target via glDrawBuffer(),
	// than to attach a different texture
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, attachmentpoints[writeTex],
							   GL_TEXTURE_2D, yTexID[writeTex], 0);
	checkGLErrors("perform 0");
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, attachmentpoints[readTex],
							   GL_TEXTURE_2D, yTexID[readTex], 0);
	checkGLErrors("perform 1");

	// check if that worked
    if ( !checkFramebufferStatus() )
	{
		fputs( "glFramebufferTexture2DEXT() FAILED!\n", stderr );
		exit(1);
	}

	// enable texture x (read-only, not changed in the computation loop)
	glActiveTexture( GL_TEXTURE1 );	
	glBindTexture( GL_TEXTURE_2D, xTexID );
	glUniform1i( xUni, 1 );								// tex unit 1

	// write scalar alpha into shader's uniform variable
	glUniform1f( alphaUni, alpha );

	// Calling glFinish() is only neccessary to get accurate timings,
	// and we need a high number of iterations to avoid timing noise.
    glFinish();

    double start = clock();
    for ( int i = 0; i < numIterations; i ++ )
	{
		// set render destination
		glDrawBuffer( attachmentpoints[writeTex] );

		// enable texture y_old (read-only)
		glActiveTexture( GL_TEXTURE0 );
		glBindTexture( GL_TEXTURE_2D, yTexID[readTex] );
		glUniform1i( yUni, 0 );							// texunit 0

		// now render multitextured viewport-sized quad
		
		// make quad filled to hit every pixel/texel 
		// (should be default but we never know)
		glPolygonMode( GL_FRONT, GL_FILL );

		// and render the quad with normalized texcoords
		glBegin(GL_QUADS);
			glTexCoord2f( 0.0, 0.0 ); glVertex2f( 0.0,      0.0 );
			glTexCoord2f( 1.0, 0.0 ); glVertex2f( texSizeX, 0.0 );
			glTexCoord2f( 1.0, 1.0 ); glVertex2f( texSizeX, texSizeY );
			glTexCoord2f( 0.0, 1.0 ); glVertex2f( 0.0,      texSizeY );
		glEnd();

		// swap role of the two textures (read-only source becomes 
		// write-only target and the other way round):
		swapTextures();
    }

	// done, stop timer, calc MFLOP/s if neccessary
	glFinish();

	double end = clock();
	double total = (end-start)/CLOCKS_PER_SEC;
	double mflops = (2.0*N*numIterations) / (total * 1000000.0);
	printf( "MFLOPs/sec for N=%d:  %d\n", N, static_cast<int>(mflops) );

	// done, just do some checks if everything went smoothly.
    checkFramebufferStatus();
    checkGLErrors("performComputation");
}



/**
 * Checks for OpenGL errors.
 * Extremely useful debugging function: When developing, 
 * make sure to call this after almost every GL call.
 */
void checkGLErrors( const char *label )
{
    GLenum errCode = glGetError();
    if ( errCode == GL_NO_ERROR )
		return;

	fputs( "OpenGL ERROR: ", stderr);
	fputs( (char*)gluErrorString(errCode), stderr);
	fputs( " (label: ", stderr);
	fputs( label, stderr);
	fputs( ")\n", stderr);
}


/**
 * Checks framebuffer status.
 * Copied directly out of the spec, modified to deliver a return value.
 */
bool checkFramebufferStatus(void)
{
    GLenum status;
    status = (GLenum) glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    switch(status) {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
            return true;
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
			printf("Framebuffer incomplete, incomplete attachment\n");
            return false;
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
			printf("Unsupported framebuffer format\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
			printf("Framebuffer incomplete, missing attachment\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
			printf("Framebuffer incomplete, attached images must have same dimensions\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
			printf("Framebuffer incomplete, attached images must have same format\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
			printf("Framebuffer incomplete, missing draw buffer\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
			printf("Framebuffer incomplete, missing read buffer\n");
            return false;
    }
	return false;
}



/**
 * swaps the role of the two y-textures (read-only and write-only)
 * Think of "pointer swapping".
 */
void swapTextures(void)
{
	int h = writeTex;
	writeTex = readTex;
	readTex = h;
}



// rendering

void changeSize(int w, int h)
{
	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if ( h == 0 )
		h = 1;

	windowWidth = w;
	windowHeight = h;

    glViewport( 0, 0, windowWidth, windowHeight );

	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluOrtho2D( 0.0, windowWidth, 0.0, windowHeight );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
}


// The following routine is only for visualizing the result!
// Thus, we can safely make a larger window
void renderScene(void)
{
	// no clear necessary, since we draw the whole screen, and without z
	glMatrixMode( GL_MODELVIEW );

	glLoadIdentity();
	// default for camera = 0,0,0; looking at -z; near = -1; far = +1

	// restore render destination to regular frame buffer
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );

	glActiveTexture( GL_TEXTURE0 );
	glEnable( GL_TEXTURE_2D );
	glBindTexture( GL_TEXTURE_2D, yTexID[readTex] );

	// texture states; just to make sure
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glColor3f(1,1,0);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0); 
		glVertex2f(0.0, 0.0);
		glTexCoord2f(1.0, 0.0); 
		glVertex2f(windowWidth, 0.0);
		glTexCoord2f(1.0, 1.0); 
		glVertex2f(windowWidth, windowHeight);
		glTexCoord2f(0.0, 1.0); 
		glVertex2f(0.0, windowHeight);
	glEnd();

	// no swap buffers here, because we don't animate anything
	glFinish();
}


void processNormalKeys(unsigned char key, int x, int y)
{
	if (key == 27) 
		exit(0);
}

