/*
 * Mandelbrot shader using GPGPU techniques
 *
 * Invokation: ./mandelbrot -s 256 -n 200 (for instance)
 *
 * Author: Gabriel Zachmann, June 2007;
 *
 * The code is derived from ../fbo_demo/saxpy.cpp
 */



// includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#if defined(__linux__)

#define GL_GLEXT_PROTOTYPES
#include <gl.h>
#include <glu.h>
#include <glut.h>

#elif defined(__APPLE__)

#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>

#endif

#include "shader_setup.h"


// prototypes
bool checkFramebufferStatus(void);
void checkGLErrors(const char *label);
void printResults(void);
void createTextures(void);
void initFBO(void);
void performComputation(void);
void setupTexture(const GLuint texID);
void swapTextures(void);
void transferFromTexture(float* data);
void transferToTexture(float* data, GLuint texID);
void renderScreenSizedQuad( void );
void changeSize(int w, int h);
void renderScene(void);
void processNormalKeys(unsigned char key, int x, int y);
void showReadTexture( void );


// command line options
unsigned int num_iter = 100;			// num of function iterations
float range_center[2] = {-0.6, 0.0};
float range_size = 3.0;					// square section of the Mandelbrot set
unsigned int texSize = 64;				// "resolution" of the range of the M-set

unsigned int windowWidth = 0, windowHeight = 0;

// flags to fine-tune application and to ease debugging
bool print_results = false;

// support of extensions
bool npot_supported;

// texture identifiers
GLuint texID[2];

// ping pong management vars
int writeTex = 0;
int readTex = 1;
GLenum attachmentpoints[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };

// GLSL vars
GLuint sh_prog_id;

// FBO identifier
GLuint fbo;


/**
 * main, just calls things in the appropriate order
 */

int main(int argc, char **argv)
{

	// Sets up GLUT, creates "window"
	// (better put: valid GL context, since the window is not really needed --
	// it's only needed with option -w)
	glutInit( &argc, argv );
	//
	// create window, if wanted
	glutInitDisplayMode( GLUT_RGB );
	glutInitWindowPosition(100,100);
	glutInitWindowSize( 300, 300 );
	glutCreateWindow("Mandelbrot");		// must be done before any other GL commands

	npot_supported = static_cast<bool>( glutExtensionSupported("GL_ARB_texture_non_power_of_two") );

	// parse command line
	for ( int i = 1; i < argc; i ++ )
	{
		if ( strcmp(argv[i],"-n") == 0 )
		{
			i ++ ;
			num_iter = atoi( argv[i] );
		}
		else
		if ( strcmp(argv[i],"-v") == 0 )
			print_results = true;
		else
		if ( strcmp(argv[i],"-s") == 0 )
		{
			i ++ ;
			texSize = atoi( argv[i] );
			if ( texSize < 2 )
				texSize = 2;
		}
		else
		if ( strcmp(argv[i],"-r") == 0 && argc-i >= 3 )
		{
			i ++ ;
			range_center[0] = atof( argv[i] );
			i ++ ;
			range_center[1] = atof( argv[i] );
			i ++ ;
			range_size = atof( argv[i] );
		}
		else
		{
			puts( "mandelbrot [options]\n"
				  "-n num     max. number of iterations\n"
				  "-s size    size of texture = resolution\n"
				  "-r c_x c_y s    range of M-set: s*s square centered around c_x,c_y\n"
				  "-v         output values of final texture\n"
				  "-h         this help\n" );
			exit(0);
		}
	}

	// check whether we can actually load textures of that size on the GPU
	glTexImage2D( GL_PROXY_TEXTURE_2D, 0, GL_RGBA32F_ARB,
				  texSize, texSize, 0, GL_RGBA, GL_FLOAT, NULL );
	GLint realwidth;
	glGetTexLevelParameteriv( GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &realwidth );
	if ( realwidth == 0 )
	{
		fprintf( stderr, "Can't load textures of size %u x %u on the graphics card!\n",
				 texSize, texSize );
		exit(1);
	}

	// check if we've got a NPOT texture, and if that is supported (if any)
	if ( texSize != static_cast<unsigned int>( exp2( round( log2(texSize) ) ) ) ||
	     texSize != static_cast<unsigned int>( exp2( round( log2(texSize) ) ) )    )
	{
		fprintf( stderr, "texture size is  %u x %u x RGBA\n", texSize, texSize );

		if ( npot_supported )
			fputs( "NPOT textures are OK.\n", stderr );
		else
		{
			fputs( "NPOT textures are NOT supported!\n", stderr );
			exit(1);
		}
	}

	initFBO();							// init off-screen framebuffer
	createTextures();					// create textures for vectors
    performComputation();				// and start computation

	// show final result, which is in the "read" texture

	glutDisplayFunc( renderScene );
	glutReshapeFunc( changeSize );
	glutKeyboardFunc( processNormalKeys );

	glUseProgram( 0 );				// use fixed-function pipeline

	// now show it
	glutMainLoop();					// this will never return

    glDeleteFramebuffersEXT( 1, &fbo );		// and clean up
    glDeleteTextures( 2, texID );

    return 0;							// exit
}


/**
 * Sets up a floating point texture with NEAREST filtering.
 * (mipmaps etc. are unsupported for floating point textures)
 */
void setupTexture( const GLuint texID )
{
	// make active and bind
	glBindTexture( GL_TEXTURE_2D, texID );

	// turn off filtering and wrap modes!
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	// define texture with floating point format
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA32F_ARB,
				  texSize, texSize, 0, GL_RGBA, GL_FLOAT, NULL );

	// check if that worked
	checkGLErrors("setupTexture");
}


/**
 * Transfers data from current "read" texture, and stores it in given array.
 */
void transferFromTexture(float* data)
{
	// version a: texture is attached
	// recommended on both NVIDIA and ATI
    glReadBuffer( attachmentpoints[readTex] );
	glReadPixels( 0, 0, texSize, texSize, GL_RGBA, GL_FLOAT, data );
}


/**
 * Transfers data to texture.
 * Check web page for detailed explanation on the difference between ATI and NVIDIA.
 */
void transferToTexture( float* data, GLuint texID )
{
	// version (a): HW-accelerated on NVIDIA
	glBindTexture( GL_TEXTURE_2D, texID);
	glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, texSize, texSize,
					 GL_RGBA, GL_FLOAT, data );

	// version (b): HW-accelerated on ATI
//	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, texID, 0);
//	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
//	glRasterPos2i(0,0);
//	glDrawPixels(texSize,texSize,GL_RGBA,GL_FLOAT,data);
//	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, 0, 0);
}


/**
 * creates textures
 */
void createTextures(void)
{
    // create textures
    glGenTextures( 2, texID );

    // set up textures
    setupTexture( texID[readTex] );
    setupTexture( texID[writeTex] );

	// set texenv mode from modulate (the default) to replace)
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );

    // check if something went completely wrong
    checkGLErrors("createTextures()");
}



/**
 * Create framebuffer object, bind it to reroute rendering operations
 * from the traditional framebuffer to the off-screen buffer
 */
void initFBO(void)
{
	// create FBO (off-screen framebuffer)
    glGenFramebuffersEXT( 1, &fbo );

    // bind offscreen framebuffer (that is, skip the window-specific render target)
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fbo );

	// create 'viewport' exactly with same dimensions as the texture(s)
    glViewport( 0, 0, texSize, texSize );

    // orthographic projection for 1:1 pixel=texture mapping
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluOrtho2D( 0.0, texSize, 0.0, texSize );
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // check if something went completely wrong
    checkGLErrors("initFBO()");
}



/**
 * Output contants of texture
 */
void printResults(void)
{
	// transfer contents of current "read" texture to host memory
	unsigned int n = 4*texSize*texSize;
	float * data = new float[n];
	transferFromTexture( data );

	puts("GPU results (z_n_x, z_n_y, num.iter.):");

	for ( unsigned int i = 0; i < n; i += 4 )
		printf("%10f , %10f   - %g\n", data[i], data[i+1], data[i+2] );

	delete [] data;
}

/**
 * Performs the actual calculation.
 */
void performComputation(void)
{
	// attach two textures to FBO
	// because it's much faster to select a new render target via glDrawBuffer(),
	// than to attach a different texture
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, attachmentpoints[writeTex],
							   GL_TEXTURE_2D, texID[writeTex], 0);
	checkGLErrors("perform 0");
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, attachmentpoints[readTex],
							   GL_TEXTURE_2D, texID[readTex], 0);
	checkGLErrors("perform 1");

	// check if that worked
    if ( !checkFramebufferStatus() )
	{
		fputs( "glFramebufferTexture2DEXT() FAILED!\n", stderr );
		exit(1);
	}

	// 1st phase: init array with z_1 = c
	sh_prog_id = setShaders( "mandelbrot1.vert", "mandelbrot1.frag" );

    GLint range_center_uni = glGetUniformLocation( sh_prog_id, "RangeCenter" );
	glUniform2f( range_center_uni, range_center[0], range_center[1] );
    GLint range_size_uni = glGetUniformLocation( sh_prog_id, "RangeSize" );
	glUniform1f( range_size_uni, range_size );

	// Calling glFinish() is only neccessary to get accurate timings,
	// and we need a high number of iterations to avoid timing noise.
    glFinish();

    float start = clock();

	// fill texture with initial values (z_1)
	renderScreenSizedQuad();

	// swap role of the two textures (read-only source becomes
	// write-only target and the other way round):
	swapTextures();

	if ( print_results )
	{
		puts("Texture contents before iteration");
		printResults();
	}

	// showReadTexture();				// for debugging

	// 2nd phase: do the iteration to compute z_i
	sh_prog_id = setShaders( "mandelbrot2.vert", "mandelbrot2.frag" );

    GLint zi_minus_1_uni = glGetUniformLocation( sh_prog_id, "zi_minus_1" );
	glUniform1i( zi_minus_1_uni, 0 );							// tex unit 0
    range_center_uni = glGetUniformLocation( sh_prog_id, "RangeCenter" );
	glUniform2f( range_center_uni, range_center[0], range_center[1] );
    range_size_uni = glGetUniformLocation( sh_prog_id, "RangeSize" );
	glUniform1f( range_size_uni, range_size );
    GLint cur_iteration_uni = glGetUniformLocation( sh_prog_id, "curIteration" );

    for ( unsigned int i = 1; i < num_iter; i ++ )
	{
		// pass in new iteration value
		glUniform1f( cur_iteration_uni, static_cast<float>(i) );

		// set render destination
		glDrawBuffer( attachmentpoints[writeTex] );

		// enable "read" texture (read-only), which contains results from last iteration
		glBindTexture( GL_TEXTURE_2D, texID[readTex] );

		// make quad filled to hit every pixel/texel
		renderScreenSizedQuad();

		// swap role of the two textures (read-only source becomes
		// write-only target and the other way round):
		swapTextures();
    }

	if ( print_results )
	{
		puts("Texture contents after iteration");
		printResults();
	}

	// showReadTexture();				// for debugging
	// or, you might switch this on to see, how the texture looks like with z_i

	// 3rd phase: compute nice colors from the iteration count stored in each texel
	sh_prog_id = setShaders( "mandelbrot3.vert", "mandelbrot3.frag" );

    GLint zn_uni = glGetUniformLocation( sh_prog_id, "zn" );
	glUniform1i( zn_uni, 0 );
    GLint max_iter_uni = glGetUniformLocation( sh_prog_id, "MaxIterations" );
	glUniform1f( max_iter_uni, static_cast<float>(num_iter) );
    GLint inner_color_uni = glGetUniformLocation( sh_prog_id, "InnerColor" );
	glUniform4f( inner_color_uni, 1.0, 0.0, 0.0, 0.0 );
    GLint outer_color_uni = glGetUniformLocation( sh_prog_id, "OuterColor" );
	glUniform4f( outer_color_uni, 0.0, 0.0, 1.0, 0.0 );
    GLint band_frequ_uni = glGetUniformLocation( sh_prog_id, "BandFrequ" );
	glUniform1f( band_frequ_uni, 0.1 );

	glDrawBuffer( attachmentpoints[writeTex] );
	glBindTexture( GL_TEXTURE_2D, texID[readTex] );
	renderScreenSizedQuad();
	swapTextures();


	// done, stop timer, calc MFLOP/s if neccessary
	glFinish();

	float end = clock();
	float total = (end-start)/CLOCKS_PER_SEC;
	printf( "Time for N=%d:  %f sec\n", texSize*texSize, total );

	// done, just do some checks if everything went smoothly.
    checkFramebufferStatus();
    checkGLErrors("performComputation");
}



/**
 * Checks for OpenGL errors.
 * Extremely useful debugging function: When developing,
 * make sure to call this after almost every GL call.
 */
void checkGLErrors( const char *label )
{
    GLenum errCode = glGetError();
    if ( errCode == GL_NO_ERROR )
		return;

	fputs( "OpenGL ERROR: ", stderr);
	fputs( (char*)gluErrorString(errCode), stderr);
	fputs( " (label: ", stderr);
	fputs( label, stderr);
	fputs( ")\n", stderr);
}


/**
 * Checks framebuffer status.
 * Copied directly out of the spec, modified to deliver a return value.
 */
bool checkFramebufferStatus(void)
{
    GLenum status;
    status = (GLenum) glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    switch(status) {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
            return true;
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
			printf("Framebuffer incomplete, incomplete attachment\n");
            return false;
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
			printf("Unsupported framebuffer format\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
			printf("Framebuffer incomplete, missing attachment\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
			printf("Framebuffer incomplete, attached images must have same dimensions\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
			printf("Framebuffer incomplete, attached images must have same format\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
			printf("Framebuffer incomplete, missing draw buffer\n");
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
			printf("Framebuffer incomplete, missing read buffer\n");
            return false;
    }
	return false;
}



/**
 * swaps the role of the two y-textures (read-only and write-only)
 * Think of "pointer swapping".
 */
void swapTextures(void)
{
	int h = writeTex;
	writeTex = readTex;
	readTex = h;
}



// rendering

void changeSize(int w, int h)
{
	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if ( h == 0 )
		h = 1;

	windowWidth = w;
	windowHeight = h;

    glViewport( 0, 0, windowWidth, windowHeight );

	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluOrtho2D( 0.0, windowWidth, 0.0, windowHeight );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
}


void renderScreenSizedQuad( void )
{
	// make quad filled to hit every pixel/texel
	// (should be default but we never know)
	glPolygonMode( GL_FRONT, GL_FILL );

	// and render the quad with normalized texcoords
	glBegin(GL_QUADS);
		glTexCoord2f( 0.0, 0.0 ); glVertex2f( 0.0,      0.0 );
		glTexCoord2f( 1.0, 0.0 ); glVertex2f( texSize, 0.0 );
		glTexCoord2f( 1.0, 1.0 ); glVertex2f( texSize, texSize );
		glTexCoord2f( 0.0, 1.0 ); glVertex2f( 0.0,      texSize );
	glEnd();
}



// The following routine is only for visualizing the result!
// Thus, we can safely make a larger window
void renderScene(void)
{
	// no clear necessary, since we draw the whole screen, and without z
	glMatrixMode( GL_MODELVIEW );

	glLoadIdentity();
	// default for camera = 0,0,0; looking at -z; near = -1; far = +1

	// restore render destination to regular frame buffer
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );

	glEnable( GL_TEXTURE_2D );
	glBindTexture( GL_TEXTURE_2D, texID[readTex] );

	// texture states; just to make sure
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
	glColor3f(1,1,0);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.0, 0.0);
		glTexCoord2f(1.0, 0.0);
		glVertex2f(windowWidth, 0.0);
		glTexCoord2f(1.0, 1.0);
		glVertex2f(windowWidth, windowHeight);
		glTexCoord2f(0.0, 1.0);
		glVertex2f(0.0, windowHeight);
	glEnd();

	// no swap buffers here, because we don't animate anything
	glFinish();
}


void processNormalKeys(unsigned char key, int x, int y)
{
	if (key == 27)
		exit(0);
}


// This is for debugging purposes:
// it shows the current contents of the "read" texture in a window;
// the only thing you can do after that is <esc>

void showReadTexture( void )
{
	glutDisplayFunc( renderScene );
	glutReshapeFunc( changeSize );
	glutKeyboardFunc( processNormalKeys );

	glUseProgram( 0 );				// use fixed-function pipeline

	// transfer contents of current "read" texture to host memory
	unsigned int n = 4*texSize*texSize;
	float * data = new float[n];
	transferFromTexture( data );

	// scale texels into [0,1]
	float maxi[4] = { -1E20, -1E20, -1E20, -1E20 }, mini[4] = { 1E20, 1E20, 1E20, 1E20 };
	for ( unsigned int i = 0; i < n; i ++ )
	{
		unsigned int j = i % 4;
		if ( data[i] > maxi[j] )
			maxi[j] = data[i];
		if ( data[i] < mini[j] )
			mini[j] = data[i];
	}
	for ( unsigned int i = 0; i < n; i ++ )
	{
		unsigned int j = i % 4;
		data[i] = (data[i] - mini[j]) / (maxi[j] - mini[j]);
	}

	transferToTexture( data, texID[readTex] );		// this texture is used by renderScene()
	delete [] data;

	// now show it
	glutMainLoop();					// this will never return
}

