/*
 * 2D simplex noise (this is basically an excerpt from noise.frag)
 *
 * Simplex noise is implemented by the functions:
 * float snoise(vec2 P)
 *
 * Author: Stefan Gustavson ITN-LiTH (stegu@itn.liu.se) 2004-12-05
 * Simplex indexing functions by Bill Licea-Kane, ATI (bill@ati.com)
 *
 * You may use, modify and redistribute this code free of charge,
 * provided that the author's names and this notice appear intact.
 */

/*
 * NOTE: there is a formal problem with the dependent texture lookups.
 * A texture coordinate of exactly 1.0 will wrap to 0.0, so strictly speaking,
 * an error occurs every 256 units of the texture domain, and the same gradient
 * is used for two adjacent noise cells. One solution is to set the texture
 * wrap mode to "CLAMP" and do the wrapping explicitly in GLSL with the "mod"
 * operator. This could also give you noise with repetition intervals other
 * than 256 without any extra cost.
 * This error is not even noticeable to the eye even if you isolate the exact
 * position in the domain where it occurs and know exactly what to look for.
 * The noise pattern is still visually correct, so I left the bug in there.
 * 
 */


/*
 * "permTexture" is a 256x256 texture that is used for both the permutations
 * and the 2D and 3D gradient lookup. For details, see the main C program.
 * 2D simplex noise uses only permTexture.
 */
uniform sampler2D permTexture;
uniform float time; // Used for texture animation

varying vec2 v_texCoord2D;
varying vec4 v_color;


/*
 * To create offsets of one texel and one half texel in the
 * texture lookup, we need to know the texture image size.
 */
#define ONE 0.00390625
#define ONEHALF 0.001953125
// The numbers above are 1/256 and 0.5/256, change accordingly
// if you change the code to use another texture size.



/*
 * 2D simplex noise. Somewhat slower but much better looking than classic noise.
 */
float snoise(vec2 P)
{

	// Skew and unskew factors are a bit hairy for 2D, so define them as constants
	// This is (sqrt(3.0)-1.0)/2.0
#define F2 0.366025403784
	// This is (3.0-sqrt(3.0))/6.0
#define G2 0.211324865405

	// Skew the (x,y) space to determine which cell of 2 simplices we're in
	float s = (P.x + P.y) * F2;			// Hairy factor for 2D skewing
	vec2 Pi = floor(P + s);
	float t = (Pi.x + Pi.y) * G2;		// Hairy factor for unskewing
	vec2 P0 = Pi - t;					// Unskew the cell origin back to (x,y) space
	Pi = Pi * ONE + ONEHALF;			// Integer part, scaled and offset for texture lookup

	vec2 Pf0 = P - P0;					// The x,y distances from the cell origin

	// For the 2D case, the simplex shape is an equilateral triangle.
	// Find out whether we are above or below the x=y diagonal to
	// determine which of the two triangles we're in.
	vec2 o1;
	if (Pf0.x > Pf0.y)
		o1 = vec2(1.0, 0.0);			// +x, +y traversal order
	else
		o1 = vec2(0.0, 1.0);			// +y, +x traversal order

	// Noise contribution from simplex origin
	vec2 grad0 = texture2D(permTexture, Pi).rg * 4.0 - 1.0;
	float t0 = 0.5 - dot(Pf0, Pf0);
	float n0;
	if (t0 < 0.0)
		n0 = 0.0;
	else
	{
		t0 *= t0;
		n0 = t0 * t0 * dot(grad0, Pf0);
	}

	// Noise contribution from middle corner
	vec2 Pf1 = Pf0 - o1 + G2;
	vec2 grad1 = texture2D(permTexture, Pi + o1*ONE).rg * 4.0 - 1.0;
	float t1 = 0.5 - dot(Pf1, Pf1);
	float n1;
	if (t1 < 0.0)
		n1 = 0.0;
	else
	{
		t1 *= t1;
		n1 = t1 * t1 * dot(grad1, Pf1);
	}

	// Noise contribution from last corner
	vec2 Pf2 = Pf0 - vec2(1.0-2.0*G2);
	vec2 grad2 = texture2D(permTexture, Pi + vec2(ONE, ONE)).rg * 4.0 - 1.0;
	float t2 = 0.5 - dot(Pf2, Pf2);
	float n2;
	if (t2 < 0.0)
		n2 = 0.0;
	else
	{
		t2 *= t2;
		n2 = t2 * t2 * dot(grad2, Pf2);
	}

	// Sum up and scale the result to cover the range [-1,1]
	return 70.0 * (n0 + n1 + n2);
}


void main( void )
{

	/* These lines test 2D simplex noise,
	 */
	float n = snoise(v_texCoord2D * 16.0 + time);
	//float n = noise(vec3(4.0 * v_texCoord3D.xyz * (2.0 + sin(0.5 * time))));
	//float n = snoise(vec3(2.0 * v_texCoord3D.xyz * (2.0 + sin(0.5 * time))));
	//float n = noise(vec4(8.0 * v_texCoord3D.xyz, 0.5 * time));
	//float n = snoise(vec4(4.0 * v_texCoord3D.xyz, 0.5 * time));

	gl_FragColor = v_color * vec4(0.5 + 0.5 * vec3(n, n, n), 1.0);
}

